#
# The Zope Irc Bot
#
# (The Stupidest Zope Product Ever Written)
#
# (C) Copyright 2000 The Internet (Aust) Pty Ltd
# ACN: 082 081 472  ABN: 83 082 081 472
# All Rights Reserved
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# Author: Andrew Milton <akm@theinternet.com.au>


import App, Globals, OFS
import string
import time

from Persistence import Persistent
from Globals import ImageFile, HTMLFile, HTML, MessageDialog, package_home
from Acquisition import Implicit
from OFS.Folder import Folder
from App.Extensions import getObject
from webdav.NullResource import NullResource
from ZPublisher.HTTPResponse import HTTPResponse

from IrcBot.ZopeBot import IrcBot, CommandEvent
from IRCRegistrar import RegistryId
from IrcUser import IrcUser, manage_addIrcUser, manage_addIrcUserForm, manage_editUserForm
from ZODB.PersistentMapping import PersistentMapping
from PQueue import PQueue

import ThreadLock

def manage_addIrcChannel(self, id, title, channel,
						 botNick, botServer, botPort,
						 botName,
						 REQUEST={}):
	""" Add an IRC Channel """

	obj = IrcChannel(id=id, title=title,
					 channel=channel,
					 botNick=botNick,
					 botServer=botServer,
					 botPort=botPort,
					 botName=botName)

	self._setObject(id, obj)
	obj=getattr(self, id)
	if hasattr(obj, 'postInitialisation'):
		obj.postInitialisation()
	
	if REQUEST:
		return self.manage_main(self, REQUEST)
	return ''

manage_addIrcChannelForm=HTMLFile('manage_addIrcChannelForm', globals())

from Sync import Synchronized

class IrcChannelMethods(Synchronized, Implicit):
	#
	# Shared Methods
	#
	def __init__(self):
		self.topicText=''
		self.topicSetter=''
		self.topicTime=0
		self.logLines=[]
		self.channelUsers=PersistentMapping()
		self.lock = lock = ThreadLock.allocate_lock()
		self.__call_method__=lock.guarded_apply
		
	def viewLogEntries(self, lines=50):
		return self.logLines[len(self.logLines)-lines:]
	
	def getChannel(self):
		return self.channel
			
	def getBotArgs(self):
		args=[ self.botServer,
			   self.botPort,
			   self.botNick,
			   self.botName ]
		return args

	def enablePlugins(self):
		for p in self.Plugins.keys():
			for P in self.objectIds(self.Plugins[p].plugin.meta_type):
				PP = getattr(self, P)
				PP.enableBotAction()

	def addLogEntry(self, nick, message):
		if not nick:
			nick=self.botNick

		# We're probably still syncing up
		if not self.channelUsers.has_key(nick):
			return
		
		self.logLines.append(
			{'nick':nick, 'host':self.channelUsers[nick]['hostmask'],
			 'message':message, 'msgTime':time.time()}
			)
		self.logLines=self.logLines
		self.lastUpdated=time.time()
		get_transaction().commit()
		self._p_changed=1

	def timeSinceLastUpdate(self):
		""" Show Time since last Update """
		if not hasattr(self, 'lastUpdated'):
			self.lastUpdated=time.time()
		return time.time()-self.lastUpdated
	
	def setTopic(self, topic):
		self.topicText=topic
		self.lastUpdated=time.time()		
		
	def getTopic(self):
		""" Get the Topic """
		return self.topicText

	def setTopicTime(self, topicTime, topicSetter):
		self.topicTime=topicTime
		self.topicSetter=topicSetter
		self.lastUpdated=time.time()
		
	def getTopicSetter(self):
		""" Get who set the topic """		
		return self.topicSetter

	def getTopicTime(self):
		""" Get the time the topic was set """
		return self.topicTime

	def nickChange(self, oldNick, newNick):
		self.channelUsers[newNick]=self.channelUsers[oldNick]
		self.channelUsers[newNick]['nick']=newNick
		del self.channelUsers[oldNick]
		self.channelUsers=self.channelUsers
		self.lastUpdated=time.time()
		self._p_changed=1

	def joinUser(self, nick, channel, username, host, userFlags):
		self.channelUsers[nick]={'nick':nick,
								 'channel':channel,
								 'hostmask':"%s@%s"%(username,host),
								 'userFlags':userFlags}
		self.channelUsers=self.channelUsers
		self.lastUpdated=time.time()

	def partUser(self, nick):
		if channelUsers.has_key(nick):
			del self.channelUsers[nick]
		self.lastUpdated=time.time()
		
	def signOffUser(self, nick):
		self.partUser(nick)
		self.lastUpdated=time.time()

	def findUser(self, rawString):
		command, args, text = rawString
		hostmask=args[0]
		nick, userhost=string.split(hostmask,'!')
		for u in self.objectValues('Irc User'):
			if u.matchHostMask(userhost):
				return u
		return None
		
	def findUserLevel(self, rawString):
		command, args, text = rawString
		hostmask=args[0]
		nick, userhost=string.split(hostmask,'!')
		for u in self.objectValues('Irc User'):
			if u.matchHostMask(userhost):
				return u.level
		return 0
			
	def addCommand(self, command, text=''):
		""" Send a Command """
		self.botCommands.put({'command':command,
							  'text':text},0) # Raise exception if we've stalled
		CommandEvent.set()
		
	def hasCommands(self):
		return self.botCommands.qsize()
		
	def getNextCommand(self):
		if self.hasCommands():
			command=self.botCommands.get(0)
			self._p_changed=1
			return command

class PIrcChannelMethods(Persistent, IrcChannelMethods):
	""" Dummy """
	
class IrcChannel(Folder):
	""" A class relating to irc channels """

	my_manage_main=Folder.manage_main
	index_html=HTMLFile('mainFrame', globals())
	clientPage=HTMLFile('index_html', globals())
	manage_main=clientPage	
	lastUpdate=HTMLFile('lastUpdate', globals())
	lastLines=HTMLFile('lastLines', globals())
	manage_editBotForm=HTMLFile('manage_editBotForm', globals())

	DialogHeader=HTMLFile('DialogHeader', globals())
	DialogFooter=HTMLFile('DialogFooter', globals())

	manage_editPluginsForm=HTMLFile('manage_editPluginsForm', globals())

	manage_editUser=IrcUser.manage_editUser
	manage_editUserForm=manage_editUserForm

	botCommands=PQueue(50)
	
	Plugins={}

	manage_addIrcUser=manage_addIrcUser
	manage_addIrcUserForm=manage_addIrcUserForm

	__ac_permissions__=(
		('View management screens', ('manage_tabs', 'manage_main',
									 'manage_editBotForm','initialisePlugins',
									 'manage_editPluginsForm',
									 'manage_editUsersForm',
									 'manage_addIrcUserForm',
									 'manage_addIrcUser')),
		('View',(''))
		)
	
	manage_options=(
		{ 'label':'Channel Status', 'action':'clientPage'},
		{ 'label':'Channel Commands', 'action':'manage_editBotForm'},
		{ 'label':'Channel Plugins', 'action':'manage_editPluginsForm'},
		{ 'label':'User Management', 'action':'manage_editUsersForm'},
		{ 'label':'Contents', 'action':'my_manage_main'},
		)

	ThreadMethods=PIrcChannelMethods()
	
	def manage_beforeDelete(self, item, container):
		print "before Delete"
		registrar=getattr(self, RegistryId)
		registrar.deregisterIRCBot(self.id)
		self.ThreadMethods.addCommand(["QUIT",], "Web Interface Shutdown")
			
	def __init__(self, id, title, channel,
				 botNick, botServer, botPort, botName):
		self.id=id
		self.title=title		
		self.channel=channel
		self.botNick=botNick
		self.botServer=botServer
		self.botPort=int(botPort)
		self.botName=botName

	def TextIndexerIDs(self):
		ids={}
		have_id=ids.has_key
		StringType=type('')

		while self is not None:
			if hasattr(self, 'objectValues'):
				for o in self.objectValues():
					if (hasattr(o,'meta_type') and o.meta_type=='Text Indexer'
						and hasattr(o,'id')):
						id=o.id
						if type(id) is not StringType: id=id()
						if not have_id(id):
							if hasattr(o,'title_and_id'): o=o.title_and_id()
							else: o=id
							ids[id]=id
			if hasattr(self, 'aq_parent'): self=self.aq_parent
			else: self=None

		ids=map(lambda item: (item[1], item[0]), ids.items())
		ids.sort()
		return ids

	def getIrcUserObject(self, nick):
		""" Get a User Object """
		if hasattr(self, nick):
			o= getattr(self, nick)
			if hasattr(o, 'meta_type'):
				if o.meta_type=='Irc User':
					return o
				
	def registerBotWithRegistrar(self):
		registrar=getattr(self, RegistryId)
		registrar.registerIRCBot(self.id, self)

		
	def initialisePlugins(self):
		""" yep """
		# Get all the plugins and go nuts with doughnuts.
		for p in self.Plugins.keys():
			name=self.Plugins[p].name
			setattr(self, 'manage_add%sPluginForm'%(name),
					getattr(self.Plugins[p],'manage_addPluginForm'))
			setattr(self, 'manage_add%sPlugin'%(name),
					getattr(self.Plugins[p],'manage_addPlugin'))
			
	def postInitialisation(self):
		if self.Plugins:
			self.initialisePlugins()
		self.registerBotWithRegistrar()

	def getNames(self):
		""" Return list of who's on """
		return self.ThreadMethods.channelUsers.keys()

	def processCommand(self, users=[], msg='', REQUEST={}):
		""" Process a command from the virtual client """

		if msg:
			self.ThreadMethods.addCommand(['PRIVMSG',self.channel], msg)
			self.ThreadMethods.addLogEntry('', msg)

		
		if users:
			if type(users)==type(''):
				users=[users,]

			if REQUEST.has_key('OpSubmit') and REQUEST['OpSubmit']:
				opstring="+"
				userstring=""
				for u in users:
					opstring=opstring+"o"
					userstring=userstring+u+' '
				if userstring:
					userstring=string.rstrip(userstring)					
					self.ThreadMethods.addCommand(['MODE', self.channel, opstring, userstring], None)
			if REQUEST.has_key('DeopSubmit') and REQUEST['DeopSubmit']:
				opstring="-"
				userstring=""
				for u in users:
					opstring=opstring+"o"
					userstring=userstring+u+' '
				if userstring:
					userstring=string.rstrip(userstring)
					self.ThreadMethods.addCommand(['MODE', self.channel, opstring, userstring], None)
			if REQUEST.has_key('VoiceSubmit') and REQUEST['VoiceSubmit']:
				opstring="+"
				userstring=""
				for u in users:
					opstring=opstring+"v"
					userstring=userstring+u+' '
				if userstring:
					userstring=string.rstrip(userstring)					
					self.ThreadMethods.addCommand(['MODE', self.channel, opstring, userstring], None)

			if REQUEST.has_key('KickSubmit') and REQUEST['KickSubmit']:
				for u in users:
					self.ThreadMethods.addCommand(['KICK', self.channel, u], "Web Kick")
				
			if REQUEST.has_key('AuthSubmit') and REQUEST['AuthSubmit']:
				REQUEST.set('users',users)
				return self.manage_addIrcUserForm(self, REQUEST)

		if REQUEST:
			return self.clientPage(self, REQUEST)
		return ''
		
	def getIrcUser(self, nick):
		""" Get details for one named user """
		if self.ThreadMethods.channelUsers.has_key(nick):
			return [self.ThreadMethods.channelUsers[nick],]

	
